class State(object):
    def __init__(self, positions, x, y, width, height):

        active_cells = []

        for y, row in enumerate(positions.splitlines()):
            for x, cell in enumerate(row.strip()):
                if cell == "o":
                    active_cells.append((x, y))

        board = [[False] * width for row in range(height)]

        for cell in active_cells:
            board[cell[1] + y][cell[0] + x] = True

        self.board = board
        self.width = width
        self.height = height

    def display(self):

        output = ""

        for y, row in enumerate(self.board):
            for x, cell in enumerate(row):
                if self.board[y][x]:
                    output += " o"
                else:
                    output += " ."
            output += "\n"

        return output



from hitchstory import GivenDefinition, GivenProperty, InfoDefinition, InfoProperty
from hitchstory import BaseEngine, validate, no_stacktrace_for
from strictyaml import CommaSeparated, Str, Int


class Engine(BaseEngine):
    # https://hitchdev.com/hitchstory/using/alpha/given
    given_definition = GivenDefinition(example_precondition=GivenProperty(Str()))

    # https://hitchdev.com/hitchstory/using/alpha/metadata
    info_definition = InfoDefinition(jiras=InfoProperty(schema=CommaSeparated(Str())))

    def __init__(self, paths):
        self.path = paths

    def set_up(self):
        pass

    # https://hitchdev.com/hitchstory/using/alpha/expected-exceptions
    @no_stacktrace_for(AssertionError)
    @validate(with_number=Int())
    def do_something(self, with_number=1):
        # https://hitchdev.com/hitchstory/using/alpha/steps-and-step-arguments
        assert with_number == 1

    def tear_down(self):
        pass



from strictyaml import load
from sys import argv, exit
from path import Path
import dirtemplate
import os


THIS_DIR = Path(__file__).realpath().abspath().dirname()


def run():
    arguments = argv[1:]
    assert len(arguments) == 1 or len(arguments) == 2
    template_type = arguments[0]
    assert template_type in ["demo", "skeleton"]
    available_templates = sorted([
        str(path.basename()) for path in THIS_DIR.joinpath(template_type).listdir()
    ])

    if len(arguments) == 1:
        template_commands = ["hk --{} {}".format(template_type, template) for template in available_templates]
        for template, template_command in zip(available_templates, template_commands):
            print("{}{}  # {}".format(
                template_command,
                (max([len(cmd) for cmd in template_commands]) - len(template_command)) * " ",
                load(THIS_DIR.joinpath(template_type, template, "hitchqs.yml").text()).data['about'],
            ))
        exit(0)

    template = arguments[1]
    if template not in available_templates:
        print("{} '{}' does not exist.".format(template_type, template))
        exit(1)

    assert template in available_templates

    template_path = THIS_DIR / template_type / template

    hitchqs_settings = load(template_path.joinpath("hitchqs.yml").text()).data

    project_path_name = hitchqs_settings["path"]

    if Path(project_path_name).exists():
        print(
            (
                "Directory '{}' already exists here, "
                "remove it to run quickstart again."
            ).format(project_path_name)
        )
        exit(1)
    cwd = Path(os.getcwd()).abspath()
    dirtemplate.DirTemplate(
        src=template_path, dest=cwd / project_path_name
    ).ignore_files("hitchqs.yml").ensure_built()
    Path(cwd / project_path_name / "fingerprint.txt").remove()
    print("Quickstart run successfully!")



from hitchstory import StoryCollection, HitchStoryException
from pathquery import pathquery
from engine import Engine
from hitchrun import DIR, expected


@expected(HitchStoryException)
def bdd(*keywords):
    """
    Run story with name containing keywords.
    """
    StoryCollection(pathquery(DIR.key).ext("story"), Engine(DIR)).shortcut(
        *keywords
    ).play()


@expected(HitchStoryException)
def regression():
    """
    Run all stories
    """
    StoryCollection(
        pathquery(DIR.key).ext("story"), Engine(DIR)
    ).ordered_by_name().play()



from commandlib import CommandError, Command, python, python_bin
from hitchrun import DIR, expected

# Usable path.py objects -- https://pathpy.readthedocs.io/en/stable/api.html

# DIR.gen -- build directory (~/.hitch/xxxxxx, where the symlink 'gen' links to)
# DIR.project -- the directory containng the "hitch" folder.
# DIR.key -- the directory this file - key.py is in.
# DIR.share -- ~/.hitch/share - build folder shared build artefacts.
# DIR.cur -- the directory "hk" was run in.


# If "expected" is used, no stacktrace will be displayed for that exception
@expected(CommandError)
def hello(argument):
    """
    Try running "hk hello world".
    """
    # https://pathpy.readthedocs.io/en/stable/api.html
    DIR.gen.joinpath("hello.txt").write_text(argument)

    # https://hitchdev.com/commandlib/
    Command("cat", "hello.txt").in_dir(DIR.gen).run()


@expected(CommandError)
def runcommand():
    """
    Run python 3 code with the hk virtualenv.
    """
    python("pythoncode.py").run()  # run python code using this project's virtualenv
    python_bin.python("pythoncode.py").run()  # equivalent



class Game(object):
    def __init__(self, state, infinite_board=True):
        self.state = state
        self.width = state.width
        self.height = state.height
        self.infinite_board = infinite_board

    def step(self, count=1):

        for generation in range(count):

            new_board = [[False] * self.width for row in range(self.height)]

            for y, row in enumerate(self.state.board):
                for x, cell in enumerate(row):
                    neighbours = self.neighbours(x, y)
                    previous_state = self.state.board[y][x]
                    should_live = neighbours == 3 or (
                        neighbours == 2 and previous_state
                    )
                    new_board[y][x] = should_live

            self.state.board = new_board

    def neighbours(self, x, y):

        count = 0

        for hor in [-1, 0, 1]:
            for ver in [-1, 0, 1]:
                if not hor == ver == 0 and (
                    self.infinite_board
                    or (0 <= x + hor < self.width and 0 <= y + ver < self.height)
                ):
                    count += self.state.board[(y + ver) % self.height][
                        (x + hor) % self.width
                    ]

        return count

    def display(self):
        return self.state.display()



from hitchstory import BaseEngine, no_stacktrace_for
from seleniumdirector import WebDirector, SeleniumDirectorException
from selenium.common.exceptions import WebDriverException
from selenium import webdriver


class Engine(BaseEngine):
    def __init__(self, dirs):
        self.dirs = dirs

    def set_up(self):
        self._webdriver = webdriver.Chrome()
        self._director = WebDirector(
            self._webdriver, self.dirs.key / "selectors.yml", default_timeout=5
        )

    @no_stacktrace_for(SeleniumDirectorException)
    def visit(self, url):
        self._director.visit("http://localhost:8000/{}".format(url))

    @no_stacktrace_for(SeleniumDirectorException)
    def expect_page(self, name):
        self._director.wait_for_page(name)

    @no_stacktrace_for(SeleniumDirectorException)
    def fill_form(self, **textboxes):
        for name, text in textboxes.items():
            self._director.the(name).element.clear()
            self._director.the(name).send_keys(text)

    @no_stacktrace_for(SeleniumDirectorException)
    def click(self, item):
        self._director.the(item).click()

    @no_stacktrace_for(WebDriverException)
    def refresh_page(self):
        self._webdriver.refresh()

    def tear_down(self):
        if hasattr(self, "_webdriver"):
            self._webdriver.quit()



from hitchstory import exceptions, StoryCollection
from pathquery import pathquery
from engine import Engine
from hitchrun import DIR, expected


def _stories(engine):
    return StoryCollection(pathquery(DIR.project / "story").ext("story"), engine)


@expected(exceptions.HitchStoryException)
def bdd(*keywords):
    """
    Run story with name containing keywords.
    """
    _stories(Engine(DIR)).shortcut(*keywords).play()


@expected(exceptions.HitchStoryException)
def rbdd(*keywords):
    """
    Run story with name containing keywords and rewrite.
    """
    _stories(Engine(DIR, rewrite=True)).shortcut(*keywords).play()


@expected(exceptions.HitchStoryException)
def regression():
    """
    Run all stories.
    """
    _stories(Engine(DIR)).ordered_by_name().play()



from hitchstory import BaseEngine, GivenDefinition, GivenProperty
from strictyaml import Str


class Engine(BaseEngine):
    given_definition = GivenDefinition(website=GivenProperty(Str()))

    def __init__(self, paths):
        self.path = paths

    def set_up(self):
        print("set up")
        print(self.given["website"])

    def form_filled(self, **textboxes):
        for name, contents in sorted(textboxes.items()):
            print("{}: {}".format(name, contents))

    def clicked(self, name):
        print(name)

    def email_was_sent(self):
        print("email was sent")

    def on_failure(self, result):
        print("failure")

    def on_success(self):
        print("success")

    def tear_down(self):
        print("tear down run")



from hitchstory import BaseEngine, GivenDefinition, GivenProperty
from hitchstory import no_stacktrace_for
from hitchrunpy import ExamplePythonCode, HitchRunPyException
from templex import Templex
from strictyaml import Str
import hitchbuildpy


class Engine(BaseEngine):
    given_definition = GivenDefinition(
        setup=GivenProperty(Str()), state=GivenProperty(Str())
    )

    def __init__(self, paths, rewrite=False):
        self.path = paths
        self._rewrite = rewrite

    def set_up(self):
        virtualenv = (
            hitchbuildpy.VirtualenvBuild(
                hitchbuildpy.PyenvBuild("3.7.0").with_build_path(self.path.share),
                name="venv",
            )
            .with_requirementstxt("requirements.txt")
            .with_build_path(self.path.gen)
        )
        virtualenv.ensure_built()

        self.example_py_code = (
            ExamplePythonCode(virtualenv.bin.python, self.path.gen)
            .with_setup_code(self.given.get("setup", ""))
            .with_terminal_size(160, 100)
            .with_strings(state=self.given["state"])
            .with_modules("game.py", "state.py")
        )

    @no_stacktrace_for(HitchRunPyException)
    def run(self, code, will_output=None):
        result = self.example_py_code.with_code(code).run()

        if will_output is not None:
            try:
                Templex(will_output).assert_match(result.output)
            except AssertionError:
                if self._rewrite:
                    self.current_step.update(**{"will output": result.output})
                else:
                    raise
